// This file is part of MultiVnc-Qt.  MultiVnc-Qt is copyright 2012-2022
// by Robert W. Babcock and WSS Division of DDC and is distributed under
// the terms of the GNU General Public License.  Contact the author at
// wssddc@wssddc.com or see the project web page at
// <https://www.wssddc.com/multivnc/>.
//
// MultiVnc-Qt uses the open source Qt toolkit from the Qt Project
// (qt-project.org) and The Qt Company (www.qt.io).
//
// MultiVnc-Qt is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// MultiVnc-Qt is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MultiVnc-Qt.  If not, see <https://www.gnu.org/licenses/>.

#define EXTERN
#include "textviewer.h"
#include "ui_textviewer.h"
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "ui_pingresult.h"
#include "configuration.h"
#include "pingresult.h"
#include "pingtask.h"
#include "version.h"

//#include <specstrings.h>
//#include <winsock2.h>
//#include <iphlpapi.h>
//#ifdef __GNUC__
//	#include <icmpapi.h>
//#else
//	#include "wininclude\icmpapi.h"
//#endif

#include <QMessageBox>
#include <QPushButton>
#include <QtDebug>
#include <QHostAddress>
#include <QUdpSocket>
#include <QHostInfo>
#include <QClipboard>
#include <QRegExp>
#include <QSettings>
#include <QFile>
#include <QKeySequence>
#include <QFileInfo>
#include <QMimeData>
#include <QTimer>
#include <QElapsedTimer>
#include <QDesktopServices>
#include <QDesktopWidget>
#include <QTextEdit>
#include <QAudio>
#include <QCompleter>
#include <QDebug>
#include <QStringListModel>
#include <QtEndian>
#include <QRegularExpressionMatch>
#include <QHostAddress>
#include <QThread>
#include <QSizePolicy>
#include <QNetworkInterface>
#include <QDir>
#include <QNetworkAccessManager>
#include <QNetworkProxyFactory>

MainWindow::MainWindow(QWidget *parent) : QMainWindow(parent),
	ui(new Ui::MainWindow)
	{
	mwptr = this;
	ui->setupUi(this);

	defaultOS = 0;
	osList = QString(OSNAMES).split(",");
//	osList << OS_NAME_0 << OS_NAME_1 << OS_NAME_2 << OS_NAME_3 << OS_NAME_4 << OS_NAME_5 << OS_NAME_6 << OS_NAME_7;
	bing = new QSoundEffect;
// Sound file came from https://soundbible.com/2084-Glass-Ping.html
// License is Attr-Noncommercial 3.0
// Converted to 8-bit mono with Audacity to reduce file size
	bing->setSource(QUrl("qrc:/glass_ping-8bit-mono.wav"));
	bing->setVolume(1.0);
	configScreen = new configuration(nullptr);
	globalEnabled = true;
	soundEnabled = false;
	deletionInProgress = false;
	this->setAcceptDrops(true);

// Darken group boxes and move text right so top left corner of box is visible
	qApp->setStyleSheet(
		"QGroupBox"
			"{"
			"border: 1px solid #bbbbbb;"
			"margin-top: 2ex; padding-top: 2ex;"
			"}"
		"QGroupBox::title"
			"{"
			"margin: 0 5 0 5 px;"
			"subcontrol-origin: margin;"
			"subcontrol-position: top left;"
			"position: relative;"
			"left: 5 px;"
			"}"
		);
	ui->centralWidget->setStyleSheet("QWidget#centralWidget { background-color: #005500; }");
	QDesktopServices::setUrlHandler("slotCaller", this, "slot_slotCaller");

	glayout = new QGridLayout;
	glayout->setVerticalSpacing(0);
	createActions();
	createMenus();
	ui->statusBar->hide();

// Initialize Winsock
	WSADATA wsaData;
	int iResult = WSAStartup(MAKEWORD(2, 2), &wsaData);
	if (iResult != 0)
		{
		QMessageBox::critical(this, tr("Error"), tr("WSAStartup failed: %1").arg(iResult));
		exit(1);
		}
	ui->centralWidget->setLayout(glayout);

	keystrokeDelay = 10;
	activateDelay = 0;
	welcomePositionDelay = 200;
	welcomeSelectDelay = 200;
	logoffCadDelay = 2500;
	lockingCadDelay = 2500;
	unlockingCadDelay = 2500;
	loginCadDelay = 100;
	loginNameDelay = 0;
	warningDismissDelay = 0;
	globalWarn = WARN_YES;

	restoreSettings();

	populateHistory();
	slot_addNewHost();
	this->resize(0, 0);
	this->show();
	minWidth = this->width();
// Force width to always be enough to show all menu items
	this->setMinimumWidth(minWidth);
	forceOnscreen(this);
	CoInitializeSecurityCalled = false;
	}

MainWindow::~MainWindow()
	{
	pausePinging();
	delete bing;
	delete configScreen;
	delete ui;
	}

// -------------------- public getters and setters ---------------------

QString MainWindow::getVNCpath() { return QDir::toNativeSeparators(VNCpath); }
void MainWindow::setVNCpath(QString path) { VNCpath = path; }
QString MainWindow::getHWlistPath() {return QDir::toNativeSeparators(HWlistPath); }
void MainWindow::setHWlistPath(QString path) {HWlistPath = path; }
QString MainWindow::getMSLogonName() {return MSLogonName; }
void MainWindow::setMSLogonName(QString name) { MSLogonName = name; }
QString MainWindow::getVNCScale() { return VNCScale; }
void MainWindow::setVNCScale(QString scale) {VNCScale = scale; }
QString MainWindow::getMSLogonPassword() { return MSLogonPassword; }
void MainWindow::setMSLogonPassword(QString pw) { MSLogonPassword = pw; }
QString MainWindow::getVNCpassword() { return VNCpassword; }
void MainWindow::setVNCpassword(QString pw) { VNCpassword = pw; }
QString MainWindow::getWelcomePassword() { return welcomePassword; }
void MainWindow::setWelcomePassword(QString pw) { welcomePassword = pw; }

int MainWindow::getPingCycleTime() { return pingCycleTime; }
void MainWindow::setPingCycleTime(int time) { pingCycleTime = time; }
int MainWindow::getPingWaitTime() { return pingWaitTime; }
void MainWindow::setPingWaitTime(int time) { pingWaitTime = time; }
int MainWindow::getMaxRows() {return maxRows; }
void MainWindow::setMaxRows(int maxr) { maxRows = maxr; }
int MainWindow::getMaxCols() { return maxCols; }
void MainWindow::setMaxCols(int maxc) { maxCols = maxc; }
int MainWindow::getNgraph() { return ngraph; }
void MainWindow::setNgraph(int ngr) { ngraph = ngr; }
int MainWindow::getMaxSession() { return maxSession; }
void MainWindow::setMaxSession(int nses) { maxSession = nses; }

bool MainWindow::getViewOnlySmall() { return viewOnlySmall; }
void MainWindow::setViewOnlySmall(bool view) { viewOnlySmall = view; }
bool MainWindow::getDefaultMSLogon() { return defaultMSLogon; }
void MainWindow::setDefaultMSLogon(bool logonsetting) { defaultMSLogon = logonsetting; }
bool MainWindow::getAllowDuplicates() { return allowDuplicates; }
void MainWindow::setAllowDuplicates(bool dupe) { allowDuplicates = dupe; }
bool MainWindow::getHidePasswords() { return hidePasswords; }
void MainWindow::setHidePasswords(bool hide) { hidePasswords = hide; }
int MainWindow::getDefaultOS() { return defaultOS; }
void MainWindow::setDefaultOS(int defaultval) { defaultOS = defaultval; }

// ------------------------- private functions -------------------------

// Return true if add worked
bool MainWindow::addTarget(QString host)
	{
	host = host.trimmed().toLower();
	if(host.contains(" "))
		return false;
// If host is empty, add a new slot, otherwise,
// if there's an empty slot, find and use it
	bool reuseSlot = false;
	int taskno = pingTasks.count();
	if(!host.isEmpty())
		{
		for(int i=0; i<taskno; i++)
			{
			if(pingResults[i]->getHost().isEmpty())
				{
				taskno = i;
				reuseSlot = true;
				break;
				}
			}
		}
	if(reuseSlot)
		{
		pingTasks[taskno]->setIPvalidity(getIP(host, pingResults[taskno], pingTasks[taskno]));
		}
	else
		{
// Not reusing a slot, so check if threre is room to add one.
// Buttons at top take up the first slot in the left column
		int maxTargets = maxRows*maxCols;
		if(pingResults.count() >= maxTargets)
			{
			QMessageBox::warning(this, tr("Too many targets"),
				tr("<qt>Limit of %1 targets reached.  Increase Max rows or columns to allow more.").arg(maxTargets));
			return false;
			}

		PingResult *pr = new PingResult(this, taskno);
		PingTask *pt = new PingTask(this, taskno);
		pr->ui->lineEdit_host->setCompleter(pastHostsCompleter);
		pingTasks.append(pt);
		pingResults.append(pr);
		pt->setIPvalidity(getIP(host, pr, pt));

		connect(pr, SIGNAL(newHostSignal(int)), this, SLOT(slot_newHost(int)));
		connect(pr, SIGNAL(delHostSignal(int)), this, SLOT(slot_delHost(int)));
		connect(pr, SIGNAL(signalWOLsent(int)), this, SLOT(slot_sendWOL(int)));

		int taskCol = taskno / maxRows;
		int taskRow = taskno + 1 - taskCol * maxRows;
		glayout->addWidget(pr, taskRow, taskCol);
		QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
		this->adjustSize();
//		if(this->width() < minWidth * (taskCol+1))
//			this->resize(minWidth * (taskCol+1), this->height());

		pt->setPingTiming(pingCycleTime, pingWaitTime);
		pt->startPing();
		pr->setLastUpTime(QDateTime::currentDateTime());
		}

	loadHostSettings(host, taskno);
	return true;
	}

// Copy targets in the requested up/down/all subset to the clipboard
void MainWindow::clipTarget(deleteGroup wantedState)
	{
	QString newClipText;
	for(int i=0; i<pingResults.count(); i++)
		{
		targetState targetUp = pingResults[i]->isUp();
		switch(wantedState)
			{
			case TARGET_ALL:
				newClipText.append(pingResults[i]->getHost());
				newClipText.append("\n");
				break;
			case TARGET_DOWN:
				if(STATE_DOWN == targetUp)
					{
					newClipText.append(pingResults[i]->getHost());
					newClipText.append("\n");
					}
				break;
			case TARGET_UP:
				if(STATE_UP == targetUp)
					{
					newClipText.append(pingResults[i]->getHost());
					newClipText.append("\n");
					}
				break;
			}
		}
	if(!newClipText.isEmpty())
		{
		QClipboard *clipboard = QGuiApplication::clipboard();
		clipboard->clear();
		clipboard->setText(newClipText);
		}
	}

void MainWindow::createActions()
	{
// File Menu
	importAct = new QAction(tr("&Import"), this);
	connect(importAct, &QAction::triggered, this, &MainWindow::slot_importTargetsBrowse);

	pasteAct = new QAction(tr("&Paste"), this);
// If focus is on host field, ^V pastes there which is not right for multiple targets
	pasteAct->setShortcut(QKeySequence(Qt::ALT + Qt::Key_V));
	connect(pasteAct, &QAction::triggered, this, &MainWindow::slot_pasteTargets);

	connHelpAct = new QAction(tr("&Connection error help"), this);
	connect(connHelpAct, &QAction::triggered, this, &MainWindow::wmicHelp);

	aboutAct = new QAction(tr("&About"), this);
	connect(aboutAct, &QAction::triggered, this, &MainWindow::aboutMVNC);

	aboutQtAct = new QAction(tr("About &Qt"), this);
	connect(aboutQtAct, &QAction::triggered, this, &MainWindow::slot_aboutQt);

	verCheckAct = new QAction(tr("Check for latest version"), this);
	connect(verCheckAct, &QAction::triggered, this, &MainWindow::slot_verCheck);

	configAct = new QAction(tr("Configure"), this);
	connect(configAct, &QAction::triggered, this, &MainWindow::slot_configClicked);

	exitAct = new QAction(tr("E&xit"), this);
	connect(exitAct, &QAction::triggered, this, &MainWindow::slot_exit);

// Add menu
	addAct = new QAction(tr("&Add"));
	connect(addAct, &QAction::triggered, this, &MainWindow::slot_addNewHost);

// Clear menu
	clearAllAct = new QAction(tr("&Clear all"), this);
	connect(clearAllAct, &QAction::triggered, this, &MainWindow::slot_clearAll);

	clearDownAct = new QAction(tr("Clear &down"), this);
	connect(clearDownAct, &QAction::triggered, this, &MainWindow::slot_clearDown);

	clearUpAct = new QAction(tr("Clear &up"), this);
	connect(clearUpAct, &QAction::triggered, this, &MainWindow::slot_clearUp);

	clearLoggedinAct = new QAction(tr("Clear if anyone logged in"), this);
	connect(clearLoggedinAct, &QAction::triggered, this, &MainWindow::slot_clearLoggedIn);

// Clipboad menu
	clipAllAct = new QAction(tr("Copy &all to clipboard"), this);
	connect(clipAllAct, &QAction::triggered, this, &MainWindow::slot_clipAll);

	clipDownAct = new QAction(tr("Copy &down to clipboard"), this);
	connect(clipDownAct, &QAction::triggered, this, &MainWindow::slot_clipDown);

	clipUpAct = new QAction(tr("Copy &up to clipboard"), this);
	connect(clipUpAct, &QAction::triggered, this, &MainWindow::slot_clipUp);

// Multi actions menu
	wakeAllAct = new QAction(tr("&Wake all"), this);
	connect(wakeAllAct, &QAction::triggered, this, &MainWindow::slot_wakeAll);

	logcheckAllAct = new QAction(tr("Get logged on users all hosts"), this);
	connect(logcheckAllAct, &QAction::triggered, this, &MainWindow::slot_logcheckAll);

	upTimeAllAct = new QAction(tr("Get last boot time all hosts"), this);
	connect(upTimeAllAct, &QAction::triggered, this, &MainWindow::slot_upTimeAll);

	closeAllAct = new QAction(tr("&Close all windows"), this);
	connect(closeAllAct, &QAction::triggered, this, &MainWindow::slot_closeAll);

// Multi actions menu, VNC submenu
	VNCallAct = new QAction(tr("&Large"), this);
	connect(VNCallAct, &QAction::triggered, this, &MainWindow::slot_VNCall);

	VNCallTinyAct = new QAction(tr("&Small"), this);
	connect(VNCallTinyAct, &QAction::triggered, this, &MainWindow::slot_VNCallTiny);

// Multi actions menu, logon actions submenu
	loginAllAct = new QAction(tr("Login all"), this);
	connect(loginAllAct, &QAction::triggered, this, &MainWindow::slot_loginAll);

	lockAllAct = new QAction(tr("Lock all"), this);
	connect(lockAllAct, &QAction::triggered, this, &MainWindow::slot_lockAll);

	unlockAllAct = new QAction(tr("Unlock all"), this);
	connect(unlockAllAct, &QAction::triggered, this, &MainWindow::slot_unlockAll);

	logoffAllAct = new QAction(tr("Logoff all"), this);
	connect(logoffAllAct, &QAction::triggered, this, &MainWindow::slot_logoffAll);

	shutdownAllAct = new QAction(tr("Shutdown all"), this);
	connect(shutdownAllAct, &QAction::triggered, this, &MainWindow::slot_shutdownAll);

// Sound menu
	soundAct = new QAction(tr("&Sounds off"), this);
	connect(soundAct, &QAction::triggered, this, &MainWindow::slot_toggleSounds);
	soundAct->setToolTip("<qt>Play sound on ping state changes (target starts replying or misses 2 consecutive pings)."
									" This setting is not saved; you have to turn it on when needed.</qt>");

// Ping menu
	pingAct = new QAction(tr("P&ing on"), this);
	connect(pingAct, &QAction::triggered, this, &MainWindow::slot_togglePings);
	}

void MainWindow::createMenus()
	{
	menuBar()->setFocusPolicy(Qt::StrongFocus);
	fileMenu = menuBar()->addMenu(tr("&File"));
		fileMenu->setToolTipsVisible(true);
		fileMenu->addAction(importAct);
			importAct->setToolTip(tr("Target list from file"));
		fileMenu->addAction(pasteAct);
			pasteAct->setToolTip(tr("Target list from clipboard (Alt+V)"));
		fileMenu->addSeparator();
		fileMenu->addAction(connHelpAct);
		fileMenu->addAction(aboutAct);
		fileMenu->addAction(aboutQtAct);
		fileMenu->addAction(verCheckAct);
		fileMenu->addSeparator();
		fileMenu->addAction(configAct);
		fileMenu->addSeparator();
		fileMenu->addAction(exitAct);
	menuBar()->addAction(addAct);
	clearMenu = menuBar()->addMenu(tr("&Clear"));
		QString clearTip("<qt>Remove machines from ping list but don't touch any open RDP/VNC windows</qt>");
		clearMenu->setToolTipsVisible(true);
		clearMenu->addAction(clearAllAct);
			clearAllAct->setToolTip(clearTip);
		clearMenu->addAction(clearDownAct);
			clearDownAct->setToolTip(clearTip);
		clearMenu->addAction(clearUpAct);
			clearUpAct->setToolTip(clearTip);
		clearMenu->addAction(clearLoggedinAct);
			clearLoggedinAct->setToolTip("<qt>Remove from ping list if anyone is logged in or if query for logged in users fails</qt>");
	clipMenu = menuBar()->addMenu(tr("Clip&board"));
		clipMenu->setToolTipsVisible(true);
		clipMenu->addAction(clipAllAct);
			clipAllAct->setToolTip(tr("All target names to the clipboard"));
		clipMenu->addAction(clipDownAct);
			clipDownAct->setToolTip(tr("Down target names to the clipboard"));
		clipMenu->addAction(clipUpAct);
			clipUpAct->setToolTip(tr("Up target names to the clipboard"));
	multiMenu = menuBar()->addMenu(tr("&Multi actions"));
		multiMenu->setToolTipsVisible(true);
		multiMenu->addAction(wakeAllAct);
		multiMenu->addAction(logcheckAllAct);
			logcheckAllAct->setToolTip(tr("This will fail for localhost"));
		multiMenu->addAction(upTimeAllAct);
			upTimeAllAct->setToolTip(tr("This will fail for localhost"));
		multiMenu->addAction(closeAllAct);
			closeAllAct->setToolTip("<qt>Close windows but don't log out</qt>");
		vncMenu = multiMenu->addMenu(tr("VNC connection"));
			vncMenu->setToolTipsVisible(true);
			vncMenu->addAction(VNCallAct);
				VNCallAct->setToolTip(tr("Open VNC connection to machines that are up"));
			vncMenu->addAction(VNCallTinyAct);
				VNCallTinyAct->setToolTip(tr("Open VNC connection to machines that are up"));
		logMenu = multiMenu->addMenu(tr("Logon actions"));
			logMenu->addAction(loginAllAct);
				loginAllAct->setToolTip(tr("Log in to machines that are up"));
			logMenu->addAction(lockAllAct);
				lockAllAct->setToolTip(tr("Lock all machines that are up"));
			logMenu->addAction(unlockAllAct);
				unlockAllAct->setToolTip(tr("Unlock all machines that are up"));
			logMenu->addAction(logoffAllAct);
				logoffAllAct->setToolTip(tr("Logoff machines that are up"));
			logMenu->addAction(shutdownAllAct);
				shutdownAllAct->setToolTip(tr("Shutdown machines that are up"));
		menuBar()->addAction(soundAct);
		menuBar()->addAction(pingAct);
	}

// Don't leave focus on a host name or else, when the app loses focus,
// it looks like a new name was entered and the ping plot resets.
void MainWindow::focusOffHosts()
	{
	menuBar()->setFocus();
	}

// Lookup a host name and add name and ip to results and ping tasks
bool MainWindow::getIP(QString host_name, PingResult *pr, PingTask *pt)
	{
	if(host_name.isEmpty())
		{
		pr->setIPtext("<--- Need name or IP");
		pr->clearWOLsent();
		pr->resetData();
		pr->ui->label_downTime->clear();
		return false;
		}
#define NEWCODE
#ifdef NEWCODE
	pr->clearWOLsent();
	pr->setHost(host_name);
	pr->setButtonColor(DEFAULT_COLOR);
// No need to do a lookup (which might have a slow timeout) if host_name is an IP address already
// (but reject special addresses)
	QRegularExpression ipreg("^(?:[0-9]{1,3}\\.){3}[0-9]{1,3}$");
	if(ipreg.match(host_name).hasMatch())
		{
		if(isUnwantedIP(QHostAddress(host_name)))
			{
			pr->setIPtext("Reserved or invalid");
			pr->resetData();
			return false;
			}
		pr->setIPtext(host_name);
		pt->setIP(qFromBigEndian<quint32>(QHostAddress(host_name).toIPv4Address()));
		pr->resetData();
		pr->nameIsIP = true;
		return true;
		}

	QHostInfo info = QHostInfo::fromName(host_name);
	QList<QHostAddress> listAddr= info.addresses();
	int ipsub = -1;
	for(int i=0; i<listAddr.size(); i++)
		{
		if( listAddr.at(i).protocol() == QAbstractSocket::IPv4Protocol )
			{
			ipsub = i;
			break;
			}
		}
	if(ipsub < 0)
		{
		pr->setIPtext("Lookup failed");
		beatOnSizeUntilItFits();
		return false;
		}
// Reject special addresses such as loopback, multicast, first octet=0
	if(isUnwantedIP(listAddr[ipsub]))
		{
		pr->setIPtext("Reserved or invalid");
		pr->resetData();
		return false;
		}
// Byte order is wrong for IcmpSendEcho2
	pr->setIPtext(listAddr[ipsub].toString());
	pt->setIP(qFromBigEndian<quint32>(listAddr[ipsub].toIPv4Address()));
	pr->resetData();
	pr->nameIsIP = (host_name == listAddr[ipsub].toString());
	return true;
#endif

//#define OLDCODE
#ifdef OLDCODE
// Old code using deprecated gethostbyname()
	unsigned long ipaddr;
	struct hostent *remoteHost;
	remoteHost = gethostbyname(host_name.toLatin1());
	pr->setHost(host_name);
	if(remoteHost == nullptr)
		{
		pr->setIPtext("Lookup failed");
		return false;
		}
	else if(remoteHost->h_addrtype == AF_NETBIOS)
		{
		pr->setIPtext("NETBIOS address");
		return false;
		}
	else
		{
		ipaddr = *reinterpret_cast<u_long *>(remoteHost->h_addr_list[0]);
		if(ipaddr == INADDR_NONE)
			{
			pr->setIPtext("Lookup failed");
			return false;
			}
		}
	pr->setIPtext(IPtoQString(ipaddr));
	pt->setIP(ipaddr);
	pr->resetData();
	pr->nameIsIP = (host_name == IPtoQString(ipaddr));
	return true;
#endif
	}

// Reject special addresses such as loopback, multicast, first octet=0

bool MainWindow::isUnwantedIP(QHostAddress address)
	{
	return(!address.isGlobal() || address.isLoopback() || address.isMulticast() || address.isNull());
	}

// Check if host is already being pinged.
bool MainWindow::isDupe(QString host)
	{
	for(int i=0; i<pingResults.count(); i++)
		{
		if(0 == QString::compare(host, pingResults[i]->getHost(), Qt::CaseInsensitive))
			return true;
		}
	return false;
	}

// When deleting a target, we need to wait until any pending pings are
// completed, else the buffer that receives results may no longer exist.
// So tell everyone not to start a new ping, stop all ping cycle timers
// and wait for any pings in progress to complete.

// But... If something goes wrong, like general failure errors, the callback
// for ping completed may never come.  So, if isPingInProgress is still true
// after twice the configured wait time, assume the ping is done and cleanup.
bool MainWindow::pausePinging()
	{
	bool saveEnabled = globalEnabled;
	globalEnabled = false;
	for(int i=0; i<pingTasks.count(); i++)
		pingTasks[i]->timer->stop();
	QElapsedTimer pingwaitTimer;
	pingwaitTimer.start();

// If we find an incomplete ping task, we keep checking it until we hit twice
// the timeout value.  Once we get by the problem tasks, all others will either
// have completed or also exceeded twice the timeout value.
	for(int i=0; i<pingTasks.count(); i++)
		{
		if(pingTasks[i]->isPingInProgress())
			{
			if(pingwaitTimer.elapsed() > 2*mwptr->getPingWaitTime())
				{
				pingTasks[i]->cleanupPingDone();
				}
			else
				{
				i--;
				}
			}
		QCoreApplication::processEvents();
		QThread::yieldCurrentThread();
		}
	return saveEnabled;
	}

// Restore setting that apply to the application, but not any
// that are target-dependent
void MainWindow::restoreSettings()
	{
	QSettings mvc("WSS-DDC", "MultiVNC-Qt\\Settings");
	if(mvc.contains("MainGeometryPos"))
		{
		QPoint pos = mvc.value("MainGeometryPos").toPoint();
		move(pos);
		}
	VNCpath = mvc.value("VNCpath", "").toString();
	HWlistPath = mvc.value("HWlistPath", "").toString();
	MSLogonName = mvc.value("MSLogonName", "").toString();
	MSLogonPassword = decryptString(mvc.value("MSLogonPassword", "").toString());
	welcomePassword = decryptString(mvc.value("welcomePassword", "").toString());
	VNCpassword = decryptString(mvc.value("VNCpassword", "").toString());
	VNCScale = mvc.value("VNCScale", "1/3").toString();
	pingWaitTime = mvc.value("pingWaitTime", 1000).toInt();
	if(pingWaitTime < 1000)
		pingWaitTime = 1000;
	pingCycleTime = mvc.value("pingCycleTime", 2000).toInt();
	if(100 + pingWaitTime > pingCycleTime)
		pingCycleTime = 100 + pingWaitTime;
	viewOnlySmall = mvc.value("viewOnlySmall", true).toBool();
	defaultMSLogon = mvc.value("defaultMSLogon", false).toBool();
	allowDuplicates = mvc.value("allowDuplicates", false).toBool();
	maxRows = mvc.value("maxRows", 10).toInt();
	maxCols = mvc.value("maxCols", 1).toInt();
	ngraph = mvc.value("Ngraph", 30).toInt();
	maxSession = mvc.value("maxSession", 10).toInt();
	hidePasswords = mvc.value("hidePasswords", true).toBool();
	globalWarn = static_cast<warnSetting>(mvc.value("globalWarn", WARN_YES).toInt());
// defaultOS in the registry is a text string, need to convert to index into osList.
	QString osn = mvc.value("defaultOS", 0).toString();
	defaultOS = 0;
	for(int i=0; i<nOS; i++)
		{
		if(osList[i] == osn)
			{
			defaultOS = i;
			break;
			}
		}

	QSettings timing("WSS-DDC", "MultiVNC-Qt\\Timing");
	keystrokeDelay = timing.value("keystrokeDelay", keystrokeDelay).toInt();
	activateDelay = timing.value("activateDelay", activateDelay).toInt();
	welcomePositionDelay = timing.value("welcomePositionDelay", welcomePositionDelay).toInt();
	welcomeSelectDelay = timing.value("welcomeSelectDelay", welcomeSelectDelay).toInt();
	logoffCadDelay = timing.value("logoffCadDelay", logoffCadDelay).toInt();
	lockingCadDelay = timing.value("lockingCadDelay", lockingCadDelay).toInt();
	unlockingCadDelay = timing.value("unlockingCadDelay", unlockingCadDelay).toInt();
	loginCadDelay = timing.value("loginCadDelay", loginCadDelay).toInt();
	loginNameDelay = timing.value("loginNameDelay", loginNameDelay).toInt();
	warningDismissDelay = timing.value("warningDismissDelay", warningDismissDelay).toInt();

// If VNCPath wasn't found in settings, see if we can find it in the Windows uninstall registry.
// Add to saved settings if found.
	if(!VNCpath.isEmpty())
		return;
	QSettings uninst("HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\Ultravnc2_is1", QSettings::Registry64Format);
	QString viewer = uninst.value("InstallLocation", "").toString();
	if(viewer.isEmpty())
		return;
	if(!viewer.endsWith("\\"))
		viewer += "\\";
	viewer += "vncviewer.exe";
	if(QFileInfo::exists(viewer) && QFileInfo(viewer).isFile())
		{
		VNCpath = viewer;
		mvc.setValue("VNCpath", VNCpath);
		}
	}

// Set the global ping on/off state to pingState if it doesn't already have this value
void MainWindow::resumePinging(bool pingState)
	{
	if(pingState != globalEnabled)
		slot_togglePings();
	}

// Don't save/restore size of main window - it is set as targets are added
void MainWindow::saveSettings()
	{
	QSettings mvc("WSS-DDC", "MultiVNC-Qt\\Settings");
	mvc.setValue("MainGeometryPos", pos());
	mvc.setValue("bingVolume", volume);
	}

// Display the time a target has been down
void MainWindow::showDownTime(int ntask)
	{
	qint64 downTime = pingResults[ntask]->getLastUpTime().msecsTo(QDateTime::currentDateTime())/1000;
	int downHours = static_cast<int>(downTime / 3600);
	int downMin = static_cast<int>((downTime - 3600*downHours) / 60);
	int downSec = static_cast<int>(downTime - 3600*downHours - 60 * downMin);
	pingResults[ntask]->ui->label_downTime->
		setText(QString("%1:%2:%3").arg(downHours).arg(downMin, 2, 10, QLatin1Char('0')).arg(downSec, 2, 10, QLatin1Char('0')));
	}

// Send wake-on-lan to hardware address MAC.
// MAC must be punctuated with : or -.
// This only works within the local subnet.
void MainWindow::wakeOnLan(QString MAC)
	{
	char MACAddr [6];
	char MagicPacket [102]; // Magic package for remote boot

	QStringList octets = MAC.split(QRegExp("[:-]"), Qt::SkipEmptyParts);
	if(6 != octets.count())
		{
		QMessageBox::warning(this, "WOL error", tr("%1 is not a valid MAC address").arg(MAC));
		return;
		}
	for(int i=0; i<6; i++)
		{
		bool ok;
		MACAddr[i] = static_cast<char>(octets[i].toInt(&ok, 16));
		if(!ok)
			{
			QMessageBox::warning(this, "WOL error", tr("%1 is a bad octet").arg(octets[i]));
			return;
			}
		}

// Set to hexadecimal before the magicpacket array 6 characters ff
	memset (MagicPacket, 0xff, 6);

	int packetsize = 6;
// Build MagicPacket.
	for (int i = 0; i <16; i++)
		{
		memcpy (MagicPacket + packetsize, MACAddr, 6);
		packetsize += 6;
		}
	QUdpSocket udpSocket;
// If there are multiple netwok interfaces, send WOL to all of them
// Virtual Box network interface is one way this can hapen.

// Old code that may fail with multiple interfaces
//	udpSocket.writeDatagram(MagicPacket, 102, QHostAddress("255.255.255.255"), 9);

// This code came from a post by qxoz at
// https://forum.qt.io/topic/29912/qudpsocket-multiple-network-broadcast/5
	QList<QNetworkInterface> ifaces = QNetworkInterface::allInterfaces();

	for (int i=0; i<ifaces.size(); i++)
		{
// Get all IP addresses for the current interface
		QList<QNetworkAddressEntry> addrs = ifaces[i].addressEntries();

// And for any IP address, if it is IPv4 and the interface is active, send WOL packet
		for (int j=0; j<addrs.size(); j++)
			{
			if ((addrs[j].ip().protocol() == QAbstractSocket::IPv4Protocol) && (addrs[j].broadcast().toString() != ""))
				udpSocket.writeDatagram(MagicPacket, 102, addrs[j].broadcast(), 9);
			}
		}
	}

// ------------------------------- events ------------------------------

void MainWindow::closeEvent(QCloseEvent *event)
	{
	slot_exit();
	event->accept();
	}

void MainWindow::dragEnterEvent(QDragEnterEvent *event)
	{
// accept just text/uri-list mime format
	if (event->mimeData()->hasFormat("text/uri-list"))
		event->acceptProposedAction();
	}

void MainWindow::dropEvent(QDropEvent *event)
	{
	QList<QUrl> urlList;
	QString fName;
	QFileInfo info;

	if (event->mimeData()->hasUrls())
		{
		urlList = event->mimeData()->urls(); // returns list of QUrls

// if just text was dropped, urlList is empty (size == 0)
		if ( urlList.size() > 0) // if at least one QUrl is present in list
			{
			fName = urlList[0].toLocalFile(); // convert first QUrl to local path
			info.setFile( fName ); // information about file
			if ( info.isFile() )
				slot_importTargets(fName);
			}
		}

	event->acceptProposedAction();
	}

// --------------------------- public slots ----------------------------

void MainWindow::slot_configClicked()
	{
	configScreen->populateForm();
	configScreen->show();
	configScreen->raise();
	forceOnscreen(configScreen);
	}

// Ping results get sent here from the ping tasks callback routine.
// Trigger graph update and set button color.
void MainWindow::slot_processPings(int ntask, unsigned long pingTime)
	{
	if(pingResults.count() <= ntask)
		return;
	pingResults[ntask]->slot_newData(pingTime);
	pingResults[ntask]->processData(soundEnabled);

	if(!globalEnabled || !pingTasks[ntask]->getIPvalidity())
		{
		pingResults[ntask]->setButtonColor(DEFAULT_COLOR);
		}
	else if(STATE_UP == pingResults[ntask]->isUp())
		{
		pingResults[ntask]->setButtonColor(UP_COLOR);
		pingResults[ntask]->setLastUpTime(QDateTime::currentDateTime());
		pingResults[ntask]->ui->label_downTime->setText("");
		}
	else if(pingResults[ntask]->wakeSent())
		{
		pingResults[ntask]->setButtonColor(WOL_COLOR);
		showDownTime(ntask);
		}
	else
		{
		pingResults[ntask]->setButtonColor(DOWN_COLOR);
		showDownTime(ntask);
		}
	}

void MainWindow::aboutMVNC()
	{
	textViewer *aboutMvncHelp = new textViewer("MultiVNC-Qt", this);
	aboutMvncHelp->setAttribute(Qt::WA_DeleteOnClose);
	aboutMvncHelp->setWindowModality(Qt::WindowModal);
	aboutMvncHelp->show();

	aboutMvncHelp->ui->tvText->append(tr(
		"MultiVNC-Qt version " MVNC_VERSION ", " MVNC_DATE
		"<p>by Robert W. Babcock and WSS Division of DDC"
		"<br>wssddc@wssddc.com"

		"<p>Visit <a href=\"" WEB_LINK "\">www.wssddc.com</a>"
		" for news, updates, source code and discussion forum."

		"<p>This program has been developed both at home and at my job (now retired) at "
		"<a href=\"https://www.cfa.harvard.edu/\">Center for Astrophysics</a>."

		"<p>MultiVNC-Qt is licensed under the <a href='slotCaller:://Gpl3'>GNU Public License version 3</a>."
		"<br>Plotting uses a slightly modified version of <a href=\"https://qcustomplot.com/\">"
		"QCustomPlot</a> which is also licensed under the GPL."

		"<p>MultiVNC-Qt is designed for system administrators or anyone managing "
		"multiple systems.  It's main functions:"
		"<ul>"
		"<li>Displays a plot of recent ping times."
		"<li>Sends Wake-on-lan."
		"<li>Opens remote sessions using Windows Remote Desktop or "
			"<a href=\"https://www.uvnc.com/index.php\">UltraVNC</a>."
		"<li>Automates open session, login, logout, lock, unlock and shutdown for multiple "
			"machines at once."
		"</ul>"

		"Some features:"
		"<ul>"
		"<li>Operations such as opening a VNC session or logging in can be performed on "
			"multiple hosts at the same time.  See options under File | Multi actions"
		"<li>Previously used targets are remembered and substring matches are presented "
			"as options in a pull-down.  Old entries can be deleted with the delete key and "
			"the full remembered list can be displayed by hitting insert."
		"<li>Queries targets systems to see if anyone is logged on.  (You wouldn't want "
			"to open an unexpected VNC session on a user, would you?)"
		"<li>Queries targets for uptime and boot time."
		"<li>Displays how long a target has been unresponsice to ping (sending wake on lan "
			"resets the timer)."
		"<li>Optional audio alert when any ping state changes.  Typical use would be waiting "
			"for a target to come up, or monitoring a target you fear may go down."
		"<li>Saved passwords are encrypted using the Windows CryptProtectData API."
		"<li>Remembers VNC/RDP choice and other options on a per-machine basis."
		"<li>RDP sessions may be a configured size or full screen."
		"<li>VNC uses a global username/password, optionally overridden with a "
			"Windows login or a custom password. VNC settings on the target determine "
			"what type of credentials need to be used."
		"<li>VNC sessions may be scaled and set view-only."
		"<li>Alows copying names of hosts that are up/down/all to clipboard."
		"</ul>"

		"Quirks"
		"<ul>"
		"<li>To position the window with Autoit, it is not sufficient to only specify the "
			"window title.  You must use a command of the form<br>"
			"&nbsp;&nbsp;&nbsp;WinMove(\"[CLASS:Qt5QWindowIcon; TITLE:MultiVNC-Qt]\", \"\", x, y)<br>"
			"or else the move will stop working once you have touched any MultiVNC menu."
		"</ul>"

		"Limitations"
		"<ul>"
		"<li>If the account running MultiVNC-Qt does not have rights on the remote, "
			"RDP users will not be seen when querying for logged on users."
		"<li>Most interaction with VNC/RDP windows is done by blindly sending keystrokes "
			"with no feedback.  So there can be failures if the target doesn't respond "
			"quickly enough, and you can force failures by taking focus away from the target "
			"by clicking on something else."
		"<li>For some actions, the keystrokes that need to be sent depend on the Windows version "
			"of the target.  A global default and individual overrides can be set, but details of "
			"how a target is configured and what group policies are applied may trigger failures."
		"<li>Encrypted passwords can only be decrypted by the user who saved them on the "
			"machine where they were saved.  If you transfer settings to another machine "
			"by exporting and importing registry values, the passwords will not be "
			"decryptable on the new machine.  A password reset will invalidate access to saved "
			"passwords; a password change should not."
		"<li>Wake-on-Lan is done by a network broadcast, so unless you've done "
			"something special with your routers, WOL will only work within the local subnet."
		"<li>The Windows RDP program doesn't accept credentials on the command line, so "
			"RDP connections will default to the current user and will prompt for a password.   "
			"<a href=\"https://www.microsoft.com/en-us/download/details.aspx?id=44989\"> Microsoft's "
			"Remote Desktop Manager</a> was a better tool for managing multiple RDP sessions, "
			"but MS discontinued it rather than fix a security vulnerability.  The Windows 10 app "
			"store has a nice client supporting multiple connections."
		"<li>Each target takes a significant amount of screen space.  The "
			"maximum number of targets is set by number of rows and columns. "
			"Adjusting these limits while more than one column of machines is "
			"displayed might not work right."
		"<li>MultiVNC-Qt has no control over where Windows opens RDP or VNC "
			"windows; they'll probably all end up in the same place."
		"<li>IPv6 hosts are not supported because the Microsoft Icmp6SendEcho2 function, which I would "
			"need to use to match the IPv4 code, is explicitly documented to not use the Timeout "
			"parameter for asynchronous calls. So, I don't know how to control the ping wait time."
		"</ul>"));

	aboutMvncHelp->scrollTop();
	}

void MainWindow::populateHistory()
	{
	QSettings mvc("WSS-DDC", "MultiVNC-Qt\\perMachine");
	QStringList oldhosts = mvc.childGroups();
//	qDebug() << oldhosts.size() << "hosts known";
//	qDebug() << oldhosts;
	pastHostsModel = nullptr;
	if(pastHostsCompleter)
		pastHostsModel =static_cast<QStringListModel *>(pastHostsCompleter->model());
	else
		pastHostsCompleter = new QCompleter();
	if(pastHostsModel==nullptr)
		 pastHostsModel = new QStringListModel();

	pastHostsModel->setStringList(oldhosts);
	pastHostsCompleter->setModel(pastHostsModel);
//	pastHostsCompleter = new QCompleter(oldhosts, this);
	pastHostsCompleter->setCaseSensitivity(Qt::CaseInsensitive);
	pastHostsCompleter->setCompletionMode(QCompleter::PopupCompletion);
	pastHostsCompleter->setFilterMode(Qt::MatchContains);
	pastHostsCompleter->setMaxVisibleItems(25);
	}

void MainWindow::wmicHelp()
	{
	textViewer *wmicHelp = new textViewer("Connection error help", this);
	wmicHelp->setAttribute(Qt::WA_DeleteOnClose);
	wmicHelp->setWindowModality(Qt::WindowModal);
	wmicHelp->show();

	wmicHelp->ui->tvText->append(tr(
		"The command line queries for logged on user or boot time are<br>"
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;wmic /node:whatever [/user:me] [/password:top_secret] ComputerSystem get UserName<br>"
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;wmic /node:whatever [/user:me] [/password:top_secret] OS Get LastBootUpTime<br>"
		"If wmic doesn't work, this program won't either.<br><br>"
		"The Windows firewall must allow the connection.  Enable the pre-defined rule<br>"
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Windows Management Instrumentation (WMI-In)<br>"
		"and disable the blocking rule<br>"
		"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;File and Printer Sharing (Spooler Service - RPC-EPMAP)<br>"
		"or change it to an allow rule."
		"<p>"
		"Querying users will not see RDP users if the local account isn't an admin on the remote computer."
		"<p>"
		"Querying last boot time for a 2019 terminal server, and possibly other cases, requires "
		"an enterprise admin account."
		"<p>"
		"Logoff by API call, which is more reliable than using keyboard stuffing, requires "
		"the remote desktop services service which is only available with Windows 7 Pro or higher. "
		"Windows 7 also requires a non-default registry setting:<br>"
		"DWORD HKLM\\SYSTEM\\CurrentControlSet\\Control\\Terminal Server\\AllowRemoteRPC = 1"));

	wmicHelp->scrollTop();
	}

// --------------------------- private slots ---------------------------
// QApplication::aboutQt() generated setgeometry errors under MinGW,
// so use the version that allows specifying a null parent.
void MainWindow::slot_aboutQt()
	{
	QMessageBox::aboutQt(nullptr);
	}

// Add a new, blank target and set the focus to the host
// field so a new name can be typed in
void MainWindow::slot_addNewHost()
	{
	addTarget(QString());
	pingResults.constLast()->ui->lineEdit_host->setFocus();
	}

// Remove all targets
void MainWindow::slot_clearAll()
	{
	bool currentPingState = pausePinging();
	for(int i=pingResults.size()-1; i>=0; i--)
		slot_delHost(i, false);
	resumePinging(currentPingState);
	}

// Remove all targets that are down
// When pinging is paused, up/down status isn't valid,
// so make list of instances to delete before pausing pings.
void MainWindow::slot_clearDown()
	{
	QList<int> downList;
	for(int i=pingResults.size()-1; i>=0; i--)
		{
		if(STATE_DOWN == pingResults[i]->isUp())
			downList.append(i);
		}
	bool currentPingState = pausePinging();
	for(int i=0; i<downList.count(); i++)
		{
		slot_delHost(downList[i], false);
		}
	resumePinging(currentPingState);
	}

// Remove all target that are up
void MainWindow::slot_clearUp()
	{
	QList<int> upList;
	for(int i=pingResults.size()-1; i>=0; i--)
		{
		if(STATE_UP == pingResults[i]->isUp())
			upList.append(i);
		}
	bool currentPingState = pausePinging();
	for(int i=0; i<upList.count(); i++)
		{
		slot_delHost(upList[i],false);
		}
	resumePinging(currentPingState);
	}

// Remove all target that have a user logged in.
// Also remove if query fails
void MainWindow::slot_clearLoggedIn()
	{
	QList<int> loggedinList;
	QString queryFail("Cleared because query failed:");
	bool someFailed = false;
	ui->statusBar->show();
	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
	for(int i=pingResults.size()-1; i>=0; i--)
		{
		if(STATE_UP == pingResults[i]->isUp())
			{
			QString result;
			ui->statusBar->showMessage("Querying " + pingResults[i]->getHost());
			int rc = isAnyoneLoggedIn(pingResults[i]->getHost(), pingResults[i]->pickUsername(), pingResults[i]->pickPW(), &result);
			QCoreApplication::processEvents();
			if(1 == rc)
				{
				someFailed = true;
				queryFail.append("\n   " + pingResults[i]->getHost());
				loggedinList.append(i);
				continue;
				}
//			if("No one is logged on" == result)
			if(result.contains("No one is logged on"))
				continue;
			loggedinList.append(i);
			}
		}
	ui->statusBar->clearMessage();
	ui->statusBar->hide();
	adjustSize();

	bool currentPingState = pausePinging();
	for(int i=0; i<loggedinList.count(); i++)
		{
		slot_delHost(loggedinList[i],false);
		}
	resumePinging(currentPingState);
	QApplication::restoreOverrideCursor();

	if(someFailed)
		QMessageBox::warning(this, tr("Query failures"), queryFail);
	}

// Copy all, down or up target host names to the clipboard
void MainWindow::slot_clipAll() { clipTarget(TARGET_ALL); }
void MainWindow::slot_clipDown() { clipTarget(TARGET_DOWN); }
void MainWindow::slot_clipUp() { clipTarget(TARGET_UP); }

// Ths gets called when a delete button is clicked
void MainWindow::slot_delHost(int taskno) { slot_delHost(taskno, true); }

// Delete a target,  If needPause is true, pause pinging first.
// Functions that do multiple deletes should have already done this,
// so don't slow the user interface even more by doing it again.
// This routine is definitely not re-entrant.  Rapid clicks on delete
// lead to it being called again before it finishes and probably crashing.
// (It's the calls to processEvents that lead to the re-entrancy problem.)
// The check on deletionInProgress narrows the vulnerable window, but doesn't
// eliminate it entirely.  QMutex only protects against calls from another
// thread, so wouldn't help here.
void MainWindow::slot_delHost(int taskno, bool needPause)
	{
	if(deletionInProgress)
		return;
	deletionInProgress = true;
	bool pingState = globalEnabled;
	if(needPause)
		pingState = pausePinging();
	pingResults[taskno]->resetData();
// Move targets in layout to not leave a hole
	glayout->removeWidget(pingResults[taskno]);
	for(int i=taskno+1; i<pingResults.count(); i++)
		{
		int newCol = (i-1) / maxRows;
		int newRow = i - newCol*maxRows;
		glayout->removeWidget(pingResults[i]);
		glayout->addWidget(pingResults[i], newRow, newCol);
		}
//	int colsDisplayed = qMax(1, (pingResults.count() + maxRows - 1) / maxRows);
//	if(this->width() < minWidth * colsDisplayed)
//		this->resize(minWidth * colsDisplayed, this->height());

	PingTask *pt = pingTasks.takeAt(taskno);
	pt->deleteLater();

	PingResult *pr = pingResults.takeAt(taskno);
	pr->deleteLater();

	for(int i=taskno; i<pingResults.count(); i++)
		{
		pingResults[i]->setTaskNum(i);
		pingTasks[i]->setTaskNum(i);
		}
// Process events so the window gets resized properly, but don't
// process user inputs else we might get called again before we finish.
// (But re-entrant calls happen anyway, hence the deletionInProgress
// stuff to try to avoid re-entrance.)
	QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);

// If all targets have been deleted, adjustSize() shrinks the window too much
// and part of the menu is lost.  So make sure the width is at least as large
// as when we started up when one (blank) target was visible.
	this->adjustSize();
//	if(this->width() < minWidth)
//		this->resize(minWidth, this->height());
	focusOffHosts();
	if(needPause)
		resumePinging(pingState);
	deletionInProgress = false;
	}

void MainWindow::slot_exit()
	{
	pausePinging();
	saveSettings();
	QApplication::closeAllWindows();
	}

// Browse for a file with a list of targets to import
void MainWindow::slot_importTargetsBrowse() { slot_importTargets(QString()); }

// If importFile is empty, browse for a file starting from the directory
// of the last-used import file.
// # is a comment marker.  Comma, space or tab marks the end of a name
// so comments can follow the name.
void MainWindow::slot_importTargets(QString importFile)
	{
	bool currentPingState = pausePinging();
	QString browse;
	QSettings mvc("WSS-DDC", "MultiVNC-Qt\\Settings");
	LastImportFile = mvc.value("LastImportFile", "").toString();
	if(importFile.isEmpty())
		{
		QStringList filters;
		filters << tr("txt files (*.txt)")
				  << tr("All files (*)");
		browse = configScreen->getFileName("List of names", LastImportFile, filters);
		if(browse.isEmpty())
			{
			resumePinging(currentPingState);
			return;
			}
		}
	else
		browse = importFile;

	QFile filelist(browse);
	if(!filelist.open((QFile::ReadOnly | QIODevice::Text)))
		{
		QMessageBox::critical(this, "Error", "Unable to open machine names file");
		resumePinging(currentPingState);
		return;
		}
	QTextStream in(&filelist);
	while(!in.atEnd())
		{
		QString line = in.readLine();
		if(line.startsWith("#"))
			continue;
		QStringList names = line.split(QRegExp("[, \t]"), Qt::SkipEmptyParts);
		if(0 == names.count())
			continue;
		QString pingname = names[0].trimmed();
		if(!pingname.isEmpty())
			{
			if(!allowDuplicates && isDupe(pingname))
				continue;
			if(!addTarget(pingname))
				break;
			}
		}
	filelist.close();
	focusOffHosts();
	resumePinging(currentPingState);

	mvc.setValue("LastImportFile", browse);
	}

// VNC connections only
void MainWindow::slot_loginAll()
	{
	if(getMSLogonName().isEmpty() || getMSLogonPassword().isEmpty())
		{
		QMessageBox::warning(this, tr("Credentials needed"), tr("Enter credentials and try again"));
		slot_configClicked();
		return;
		}

	ui->statusBar->show();
	for(int i=0; i<pingResults.count(); i++)
		{
		if(pingResults[i]->getRDPstatus())
			continue;
		if(STATE_UP == pingResults[i]->isUp())
			{
			ui->statusBar->showMessage("Logging on to " + pingResults[i]->getHost());
			pingResults[i]->slot_loginFromContextMenu();
			}
		}

	ui->statusBar->clearMessage();
	ui->statusBar->hide();
	adjustSize();
	}

// VNC connections only
void MainWindow::slot_lockAll()
	{
	ui->statusBar->show();
	for(int i=0; i<pingResults.count(); i++)
		{
		if(pingResults[i]->getRDPstatus())
			continue;
		if(STATE_UP == pingResults[i]->isUp())
			{
			ui->statusBar->showMessage("Locking " + pingResults[i]->getHost());
			pingResults[i]->slot_lockFromContextMenu();
			}
		}
	ui->statusBar->clearMessage();
	ui->statusBar->hide();
	adjustSize();
	}

// VNC connections only
void MainWindow::slot_unlockAll()
	{
// This checks that default credentials have been entered.  Custom credentials
// are checked in slot_unlockFromContextMenu.
	if(getMSLogonName().isEmpty() || getMSLogonPassword().isEmpty())
		{
		QMessageBox::warning(this, tr("Credentials needed"), tr("Enter credentials and try again"));
		slot_configClicked();
		return;
		}

	ui->statusBar->show();
	for(int i=0; i<pingResults.count(); i++)
		{
		if(pingResults[i]->getRDPstatus())
			continue;
		if(STATE_UP == pingResults[i]->isUp())
			{
			ui->statusBar->showMessage("Unlocking " + pingResults[i]->getHost());
			pingResults[i]->slot_unlockFromContextMenu();
			}
		}
	ui->statusBar->clearMessage();
	ui->statusBar->hide();
	adjustSize();
	}

// VNC connections only
void MainWindow::slot_logoffAll()
	{
	ui->statusBar->show();
	for(int i=0; i<pingResults.count(); i++)
		{
		if(pingResults[i]->getRDPstatus())
			continue;
		if(STATE_UP == pingResults[i]->isUp())
			{
			ui->statusBar->showMessage("Logging off from " + pingResults[i]->getHost());
			pingResults[i]->slot_logoffFromContextMenu();
			}
		}
	ui->statusBar->clearMessage();
	ui->statusBar->hide();
	adjustSize();
	}

// VNC connections only
void MainWindow::slot_shutdownAll()
	{
	ui->statusBar->show();
	for(int i=0; i<pingResults.count(); i++)
		{
		if(pingResults[i]->getRDPstatus())
			continue;
		if(STATE_UP == pingResults[i]->isUp())
			{
			ui->statusBar->showMessage("Shutting down " + pingResults[i]->getHost());
			pingResults[i]->slot_shutdownFromContextMenu();
			}
		}
	ui->statusBar->clearMessage();
	ui->statusBar->hide();
	adjustSize();
	}

// This is called if a new value is typed into the host field.
// Trim blanks and convert to lower case.
void MainWindow::slot_newHost(int taskno)
	{
	QString host = pingResults[taskno]->getHost().trimmed().toLower();
	pingResults[taskno]->setHost(host);
	pingTasks[taskno]->setIPvalidity(
		getIP(host, pingResults[taskno], pingTasks[taskno]));
	pingResults[taskno]->setLastUpTime(QDateTime::currentDateTime());
	loadHostSettings(host, taskno);
	}

void MainWindow::loadHostSettings(QString host, int taskno)
	{
	if(!host.isEmpty())
		{
		QSettings mvc("WSS-DDC", "MultiVNC-Qt\\perMachine\\" + host);
		int vncDef = getDefaultMSLogon() ? VNC_MS_DOMAIN : VNC_COMMON;
		pingResults[taskno]->setVNCpwType(static_cast<VNCpasswordType>(mvc.value("VNCpwType", vncDef).toInt()));
		pingResults[taskno]->setCustomVNCpassword(decryptString(mvc.value("CustomVNCpassword", "").toString()));
		pingResults[taskno]->setRDPstatus(mvc.value("useRDP", false).toBool());
		pingResults[taskno]->updateButtonLabels();
		pingResults[taskno]->setRDPsize(mvc.value("RDPsize", QPoint(900,600)).toPoint());
		QString def = mvc.value("customOS", OS_DEFAULT).toString();
		pingResults[taskno]->setCustomOS(-1);
		for(int i=0; i<nOS; i++)
			{
			if(def == osList[i])
				{
				pingResults[taskno]->setCustomOS(i);
				break;
				}
			}
		pingResults[taskno]->setSlowness(mvc.value("slowPC", false).toBool());
		pingResults[taskno]->setWarnSetting(static_cast<warnSetting>(mvc.value("warnSetting", WARN_DEFAULT).toInt()));
		pingResults[taskno]->setWelcomeStatus(mvc.value("useWelcome", false).toBool());
		pingResults[taskno]->setWelcomeUserNumber(mvc.value("welcomeUserNumber", 1).toInt());
		pingResults[taskno]->setCustomWindowsName(mvc.value("CustomWindowsName", "").toString());
		pingResults[taskno]->setCustomWindowsPW(decryptString(mvc.value("CustomWindowsPW", "").toString()));
// Just querying for values will create the host registry key if it doesn't already exist,
// so the name completer list needs to be updated.
		populateHistory();
		}
	}

// Paste a list of names from the clipboard.
// Pasting garbage may delay response as name server lookups fail.
void MainWindow::slot_pasteTargets()
	{
	QClipboard *clipboard = QGuiApplication::clipboard();
	QStringList newTargets = clipboard->text().split(QRegExp("[\r\n]"), Qt::SkipEmptyParts);
	for(int i=0; i<newTargets.count(); i++)
		{
		if(!allowDuplicates && isDupe(newTargets[i]))
			continue;
		if(!addTarget(newTargets[i]))
			break;
		}
	focusOffHosts();
	}

// Send wake-on-lan to a target if the target is not already up
void MainWindow::slot_sendWOL(int taskno)
	{
	QString wakename = pingResults[taskno]->getHost();
	if(wakename.isEmpty())
		return;
	if(STATE_UP == pingResults[taskno]->isUp())
		{
//		qDebug() << wakename << " is already awake";
		return;
		}

	if(HWlistPath.isEmpty())
		{
		QMessageBox::warning(this, "Error", "Hardware address file needs to be set.");
		slot_configClicked();
		return;
		}
	QFile hw(HWlistPath);
	if(!hw.open(QFile::ReadOnly | QIODevice::Text))
		{
		QMessageBox::critical(this, "Error", "Unable to open hardware address file");
		return;
		}
	QTextStream in(&hw);
	QString line;
	while(!in.atEnd())
		{
		line = in.readLine();
		if(line.startsWith("#"))
			continue;
		QStringList name_hw = line.split(QRegExp("[, \t]"), Qt::SkipEmptyParts);
		if(0 == name_hw.count())
			continue;
		if(name_hw[0] == wakename)
			{
			wakeOnLan(name_hw[1]);
			pingResults[taskno]->setLastUpTime(QDateTime::currentDateTime());
			pingResults[taskno]->setWOLsent();
			hw.close();
			return;
			}
		}
	hw.close();
	QMessageBox::warning(this, tr("Not found"), tr("%1 not found in hardware list").arg(wakename));
	}

void MainWindow::slot_togglePings()
	{
	globalEnabled = !globalEnabled;
	for(int i=0; i<pingTasks.size(); i++)
		{
		if(globalEnabled)
			pingTasks[i]->startPing();
		else
			{
			pingResults[i]->clearWOLsent();
			pingResults[i]->setButtonColor(DEFAULT_COLOR);
			}
		}

//	toggleButton->setText(globalEnabled ? "P&inging" : "&Idle");
	pingAct->setText(globalEnabled ? "P&inging" : "&Idle");
	}

void MainWindow::slot_toggleSounds()
	{
	soundEnabled = !soundEnabled;
//	soundButton->setText(tr("&Sound effects are %1").arg(soundEnabled ? "on" : "off"));
	soundAct->setText(tr("&Sounds %1").arg(soundEnabled ? "on" : "off"));
	}

// Open VNC for all targets (except those using RDP)
void MainWindow::slot_VNCall()
	{
	if(!sessionLimit())
		return;
	for(int i=0; i<pingResults.count(); i++)
		{
		if(STATE_UP == pingResults[i]->isUp())
			pingResults[i]->slot_VNClargeClicked();
		}
	}

// Open VNC (small window)for all targets (except those using RDP)
void MainWindow::slot_VNCallTiny()
	{
	if(!sessionLimit())
		return;
	for(int i=0; i<pingResults.count(); i++)
		{
		if(STATE_UP == pingResults[i]->isUp())
			pingResults[i]->slot_VNCsmallClicked();
		}
	}

// Count how many targets are up and return true if the count
// is less than or equal to the session limit.  Otherwise, prompt
// the user to either cancel or open more targets than the
// configured limit.  Also check if any VNC target lacks credentials.
bool MainWindow::sessionLimit()
	{
	int nup = 0;
	for(int i=0; i<pingResults.size(); i++)
		{
		if(STATE_UP ==pingResults[i]->isUp())
			{
			nup++;
			if(!pingResults[i]->getRDPstatus())
				continue;
			int pwType = pingResults[i]->getVNCpwType();
			if(VNC_DEFAULT == pwType)
				pwType = getDefaultMSLogon() ? VNC_MS_DOMAIN : VNC_COMMON;
			switch(pwType)
				{
				case VNC_COMMON:
					if(!getVNCpassword().isEmpty())
						continue;
					break;
				case VNC_MS_DOMAIN:
					if(!getMSLogonName().isEmpty() && !getMSLogonPassword().isEmpty())
						continue;
					break;
				case VNC_CUSTOM:
					if(pingResults[i]->getCustomVNCpassword().isEmpty())
						{
						QMessageBox::warning(this, tr("Credentials needed"), tr("Enter custom credentials for %1 and try again")
													.arg(pingResults[i]->getHost()));
						pingResults[i]->slot_settingsClicked();
						return false;
						}
						continue;
				}
			QMessageBox::warning(this, tr("Credentials needed"), tr("Enter credentials and try again"));
			slot_configClicked();
			return false;
			}
		}
	if(nup <= maxSession)
		return true;
	QMessageBox msgBox;
	msgBox.setText(tr("%1 targets will be opened.  Is that OK?").arg(nup));
	QPushButton *openButton = msgBox.addButton(tr("Do it"), QMessageBox::ActionRole);
	msgBox.addButton(tr("No, that's too many"), QMessageBox::ActionRole);
	msgBox.setStyleSheet("QPushButton {padding-left: 20px; padding-right: 20px;"
										 "padding-top: 5px; padding-bottom: 5px;}");

	msgBox.exec();
	if(openButton == msgBox.clickedButton())
		return true;
	return false;
	}

// Get logged on user for all hosts
void MainWindow::slot_logcheckAll()
	{
	userList.clear();
	ui->statusBar->show();
	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
	for(int i=0; i<pingResults.size(); i++)
		{
		ui->statusBar->showMessage("Querying " + pingResults[i]->getHost());
		QCoreApplication::processEvents();
		pingResults[i]->logcheck(true);
		}
	ui->statusBar->clearMessage();
	ui->statusBar->hide();
	this->adjustSize();
	QApplication::restoreOverrideCursor();

	showUserList(VIEW_LOGGEDIN);
	}

// Get boot time for all hosts
void MainWindow::slot_upTimeAll()
	{
	userList.clear();
	ui->statusBar->show();
	QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
	for(int i=0; i<pingResults.size(); i++)
		{
		ui->statusBar->showMessage("Querying " + pingResults[i]->getHost());
		QCoreApplication::processEvents();
		pingResults[i]->upcheck(true);
		}
	ui->statusBar->clearMessage();
	ui->statusBar->hide();
	this->adjustSize();
	QApplication::restoreOverrideCursor();

	showUserList(VIEW_UPTIME);
	}

void MainWindow::showUserList(viewerType vt)
	{
	bool anyred = false;
	bool currentPingState = pausePinging();
	textViewer *users = new textViewer("", this);
	users->setToolTip("<html><body>RDP users will not be seen unless your account has admin rights on the target computer.</body></html");
	users->ui->tvText->setLineWrapMode(QTextEdit::NoWrap);
	users->resize(50, 50);
	users->setAttribute(Qt::WA_DeleteOnClose);
	users->setWindowModality(Qt::WindowModal);
	users->show();

	switch(vt)
		{
		case VIEW_UPTIME:
			users->setWindowTitle("Last boot time");
			users->ui->tvText->insertHtml(tr("<u>Last boot time and uptime</u><br>"));
			break;
		case VIEW_LOGGEDIN:
			users->setWindowTitle("Logged on users");
			users->ui->tvText->insertHtml(tr("<u>Logged on users</u><br>"));
			break;
		}
	for(int i=0; i<userList.count(); i++)
		{
		users->ui->tvText->insertHtml(tr("<br>"));
		if((userList[i]).contains("\\") || userList[i].contains("No one") || userList[i].contains("booted"))
			users->ui->tvText->insertHtml(userList[i]);
		else
			{
			users->ui->tvText->insertHtml("<font color=\"red\">" + userList[i] + "</font>");
			anyred = true;
			}
		}

	if(anyred)
		{
		users->ui->tvText->insertHtml(tr("<br><br><br><a href='slotCaller:://ConHelp'>Connection failure help</a>"));
		}

	QRect screen = QApplication::desktop()->screenGeometry(this);
	int height = qMin(screen.height()*4/5, static_cast<int>(users->ui->tvText->document()->size().height())+25);
	int width = qMin(screen.width()*4/5,(static_cast<int>(users->ui->tvText->document()->size().width())+25));
	users->ui->tvText->setFixedHeight(height);
	users->ui->tvText->setFixedWidth(width);
//	users->ui->tvText->setFixedHeight(static_cast<int>(users->ui->tvText->document()->size().height())+25);
//	users->ui->tvText->setFixedWidth(static_cast<int>(users->ui->tvText->document()->size().width())+25);
//	users->resize(0, 0);
	users->setSizeSaving(false);
	resumePinging(currentPingState);
	}

// Close all RDP and VNC windows
void MainWindow::slot_closeAll()
	{
	for(int i=0; i<pingResults.size(); i++)
		pingResults[i]->slot_closeFromContextMenu();
	}

// Send wake-on-lan to all targets that are down
void MainWindow::slot_wakeAll()
	{
	for(int i=0; i<pingResults.size(); i++)
		{
		if(STATE_DOWN == pingResults[i]->isUp())
			slot_sendWOL(i);
		}
	}

// URL-like calls of format slotcaller://... come here
void MainWindow::slot_slotCaller(QUrl url)
	{
	if(0 == url.toString().compare("slotcaller:://QtAbout"))
		{
		QApplication::aboutQt();
		return;
		}
	if(0 == url.toString().compare("slotcaller:://Gpl3"))
		{
		gpl3License();
		return;
		}
	if(0 == url.toString().compare("slotcaller:://ConHelp"))
		{
		wmicHelp();
		return;
		}
	}

void MainWindow::slot_verCheck()
	{
	QNetworkAccessManager *manager = new QNetworkAccessManager(this);
// Use system proxy settings, if any
	QNetworkProxyFactory::setUseSystemConfiguration(true);
	connect(manager, SIGNAL(finished(QNetworkReply*)),
			  this, SLOT(slot_replyFinished(QNetworkReply*)));

	manager->get(QNetworkRequest(QUrl(VERSION_CHECK_URL)));
	}

void MainWindow::slot_replyFinished(QNetworkReply *reply)
	{
	if(QNetworkReply::NoError == reply->error())
		{
		QIODevice *answer = reply;
// Careful, version from web may have CR/LF appended
		QString version = answer->readAll().trimmed();
		if(0 == version.compare(MVNC_VERSION, Qt::CaseInsensitive))
			{
			QMessageBox::information(this, tr("Version check"),
				tr("You are running the latest version of MultiVnc-Qt."));
			}
		else
			{
			QMessageBox::information(this, tr("Version check"),
				tr("<html><head/><body><p>You are running MultiVnc-Qt version " MVNC_VERSION
					"<br>The latest version is %1<br><br>"
					"For updates, visit <a href='https://www.wssddc.com/MultiVnc/"
					"#Downloads'>www.wssddc.com</a>").arg(version));
			}
		}
	else
		{
		QMessageBox::warning(this, tr("Version check failed"),
			tr("Sorry, unable to check version at this time."));
		}

	reply->deleteLater();
	}

// If widget top is totally off screen, move it to (100,100).  Require
// 5 pixels visible so there's something to grab to move the window.
// Remember this is a left handed coordinate system so y=0 at the top.
// Easiest way to trigger need for this adjustment would be to have a
// saved position point to a 2nd monitor that is no longer present.
void MainWindow::forceOnscreen(QWidget *widget)
	{
// Show the widget or else geometry numbers will probably be wrong
	widget->show();
	QRect widgetscreen = QApplication::desktop()->screenGeometry(widget);
	if(widget->geometry().left() > widgetscreen.right()-5 ||
		widget->geometry().right() < widgetscreen.left()+5 ||
		widget->geometry().top() < widgetscreen.top()-5 ||
		widget->geometry().top() > widgetscreen.bottom()+5)
		{
		widget->move(QPoint(100, 100));
		}
	}

double MainWindow::getVolume()
	{
	return volume;
	}

void MainWindow::setVolume(int newvol)
	{
	volume = newvol;
	qreal linearVolume = QAudio::convertVolume(volume / qreal(100.0),
		QAudio::LogarithmicVolumeScale, QAudio::LinearVolumeScale);
	bing->setVolume(linearVolume);
	}

void MainWindow::playBing()
	{
	bing->play();
	}

// This is kind of ugly, but it's the only way I've been able to get
// sizing to work right when a long host name is entered.
void MainWindow::beatOnSizeUntilItFits()
	{
	for(int i=0; i<3; i++)
		{
		QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
		adjustSize();
		}
	}
