// This file is part of MultiVnc-Qt.  MultiVnc-Qt is copyright 2012-2022
// by Robert W. Babcock and WSS Division of DDC and is distributed under
// the terms of the GNU General Public License.  Contact the author at
// wssddc@wssddc.com or see the project web page at
// <https://www.wssddc.com/multivnc/>.
//
// MultiVnc-Qt uses the open source Qt toolkit from the Qt Project
// (qt-project.org) and The Qt Company (www.qt.io).
//
// MultiVnc-Qt is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// MultiVnc-Qt is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MultiVnc-Qt.  If not, see <https://www.gnu.org/licenses/>.

#include "mainwindow.h"

#include "Windows.h"
//MinGW doesn't support WTSOpenServerExA so a Microsoft compiler is required
#include "WtsApi32.h"
#include <winnetwk.h>

QString getLastErrorMsg();

#include <QtDebug>
#include <QStringList>

// returns 0 if successful, 1 if failed for any reason
int MainWindow::isAnyoneLoggedInByRDP(QString host, QString uname, QString pw, QString *result)
	{
	Q_UNUSED(uname)
	Q_UNUSED(pw)
	QByteArray rdphost = host.toLatin1();
	HANDLE rdph = WTSOpenServerExA(host.toLatin1().data());
//qDebug() << "Handle is" << rdph;

	DWORD pLevel = 1;
	PWTS_SESSION_INFO_1A ppSessionInfo = nullptr;
	DWORD pCount;
	if(!WTSEnumerateSessionsExA(rdph, &pLevel, 0, &ppSessionInfo, &pCount))
		{
//		qDebug() << "Error code is" << GetLastError();
//		qDebug() << "WTSEnumerateSessionsExA call failed" << getLastErrorMsg();
		WTSCloseServer(rdph);
		return 1;
		}

// Possible state values
//	0	WTSActive              User logged on to WinStation
//	1	WTSConnected           WinStation connected to client
//	2	WTSConnectQuery        In the process of connecting to client
//	3	WTSShadow              Shadowing another WinStation
//	4	WTSDisconnected        WinStation logged on without client
//	5	WTSIdle                Waiting for client to connect
//	6	WTSListen              WinStation is listening for connection
//	7	WTSReset               WinStation is being reset
//	8	WTSDown                WinStation is down due to error
//	9	WTSInit                WinStation in initialization

// Interesting cases seem to be user not blank
// State = 4 means disconnected

// Collect sessions in a QStringList so we can sort them
	int countFound = 0;
	QStringList sessions;
//qDebug() << pCount << " sessions seen";
	for(unsigned long i=0; i<pCount; i++)
		{
//		qDebug() << "Session =" << ppSessionInfo[i].pSessionName;
//		qDebug() << "  State =" << ppSessionInfo[i].State;
//		qDebug() << "   User =" << ppSessionInfo[i].pUserName;
		if(ppSessionInfo[i].pUserName)
			{
			sessions.append(QString::fromStdString(ppSessionInfo[i].pDomainName).toLower()
				+ QString("\\") + QString::fromStdString(ppSessionInfo[i].pUserName));
			if(4 == ppSessionInfo[i].State)
				sessions[countFound] += QString(" (disconnected)");
			countFound++;
			}
		}
	sessions.sort(Qt::CaseInsensitive);
#define INDENTME "<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
	if(countFound > 1)
		*result += INDENTME;
	for (int i=0; i<countFound; i++)
		{
		if(countFound > 1 && i > 0)
			*result = *result + INDENTME;
		*result += sessions[i];
		}
	WTSFreeMemoryExA(WTSTypeSessionInfoLevel1, ppSessionInfo, pCount);
	WTSCloseServer(rdph);
	return 0;
	}
