// This file is part of MultiVnc-Qt.  MultiVnc-Qt is copyright 2012-2022
// by Robert W. Babcock and WSS Division of DDC and is distributed under
// the terms of the GNU General Public License.  Contact the author at
// wssddc@wssddc.com or see the project web page at
// <https://www.wssddc.com/multivnc/>.
//
// MultiVnc-Qt uses the open source Qt toolkit from the Qt Project
// (qt-project.org) and The Qt Company (www.qt.io).
//
// MultiVnc-Qt is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, either version 3 of the License, or (at your
// option) any later version.
//
// MultiVnc-Qt is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MultiVnc-Qt.  If not, see <https://www.gnu.org/licenses/>.

#include "timingtweaks.h"
#include "ui_timingtweaks.h"
#include "mainwindow.h"

#include <QSettings>

TimingTweaks::TimingTweaks(QWidget *parent) : QMainWindow(parent), ui(new Ui::TimingTweaks)
	{
	ui->setupUi(this);
	restoreSettings();
	ui->spinBox_activateDelay->setValue(mwptr->activateDelay);
	ui->spinBox_keystrokesDelay->setValue(mwptr->keystrokeDelay);
	ui->spinBox_lockCadDelay->setValue(mwptr->lockingCadDelay);
	ui->spinBox_unlockCadDelay->setValue(mwptr->unlockingCadDelay);
	ui->spinBox_loginCadDelay->setValue(mwptr->loginCadDelay);
	ui->spinBox_logoffCadDelay->setValue(mwptr->logoffCadDelay);
	ui->spinBox_warningDismissDelay->setValue(mwptr->warningDismissDelay);
	ui->spinBox_welcomePositionDelay->setValue(mwptr->welcomePositionDelay);
	ui->spinBox_welcomeSelectDelay->setValue(mwptr->welcomeSelectDelay);

	connect(ui->pushButton_close, SIGNAL(clicked(bool)), this, SLOT(slot_closeClicked()));
	connect(ui->pushButton_saveAndClose, SIGNAL(clicked(bool)), this, SLOT(slot_saveAndCloseClicked()));
	connect(ui->pushButton_cancel, SIGNAL(clicked(bool)), this, SLOT(slot_cancelClicked()));

	ui->centralwidget->setStyleSheet("QWidget#centralwidget { background-color: #005500; }");

	this->setStyleSheet("QGroupBox { background-color:"
							  " qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,"
							  " stop: 0 #CCE5FF, stop: 1 #BBDDFF); }");
	}

TimingTweaks::~TimingTweaks()
	{
	delete ui;
	}

void TimingTweaks::slot_cancelClicked()
	{
	close();
	}

void TimingTweaks::slot_closeClicked()
	{
	useTimings();
	QSettings timing("WSS-DDC", "MultiVNC-Qt\\Timing");
	timing.setValue("TimingGeometryPos", pos());
	timing.setValue("TimingGeometrySize", size());
	close();
	}

void TimingTweaks::slot_saveAndCloseClicked()
	{
	useTimings();
	saveSettings();
	slot_closeClicked();
	}
void TimingTweaks::useTimings()
	{
	mwptr->keystrokeDelay = ui->spinBox_keystrokesDelay->value();
	mwptr->activateDelay = ui->spinBox_activateDelay->value();
	mwptr->lockingCadDelay = ui->spinBox_lockCadDelay->value();
	mwptr->loginCadDelay = ui->spinBox_loginCadDelay->value();
	mwptr->logoffCadDelay = ui->spinBox_logoffCadDelay->value();
	mwptr->unlockingCadDelay = ui->spinBox_unlockCadDelay->value();
	mwptr->warningDismissDelay = ui->spinBox_warningDismissDelay->value();
	mwptr->welcomePositionDelay = ui->spinBox_welcomePositionDelay->value();
	mwptr->welcomeSelectDelay = ui->spinBox_welcomeSelectDelay->value();
	mwptr->loginNameDelay = ui->spinBox_loginNameDelay->value();
	}

void TimingTweaks::saveSettings()
	{
	QSettings timing("WSS-DDC", "MultiVNC-Qt\\Timing");
	timing.setValue("keystrokeDelay", mwptr->keystrokeDelay);
	timing.setValue("activateDelay", mwptr->activateDelay);
	timing.setValue("welcomePositionDelay", mwptr->welcomePositionDelay);
	timing.setValue("welcomeSelectDelay", mwptr->welcomeSelectDelay);
	timing.setValue("logoffCadDelay", mwptr->logoffCadDelay);
	timing.setValue("lockingCadDelay", mwptr->lockingCadDelay);
	timing.setValue("unlockingCadDelay", mwptr->unlockingCadDelay);
	timing.setValue("loginCadDelay", mwptr->loginCadDelay);
	timing.setValue("loginNameDelay", mwptr->loginNameDelay);
	timing.setValue("warningDismissDelay", mwptr->warningDismissDelay);
	}

void TimingTweaks::restoreSettings()
	{
	QSettings timing("WSS-DDC", "MultiVNC-Qt\\Timing");
	if(timing.contains("TimingGeometryPos"))
		{
		QPoint pos = timing.value("TimingGeometryPos").toPoint();
		move(pos);
		}
	if(timing.contains("TimingGeometrySize"))
		{
		QSize size = timing.value("TimingGeometrySize").toSize();
		resize(size);
		}
	}
